c 3D PPR UEL for ABAQUS Code (Compatible with 4 node Tet Elements)
c Reference: D Spring, GH Paulino, "A Growing Library of Cohesive
c Elements for Use in ABAQUS". Engineering Fracture Mechanics
c ====================================================================
      SUBROUTINE UEL (RHS, AMATRX, SVARS, ENERGY, NDOFEL, NRHS, 
     1 NSVARS, PROPS, NPROPS, COORDS, MCRD, NNODE, U, DU, V, A, JTYPE,
     2 TIME, DTIME, KSTEP, KINC, JELEM, PARAMS, NDLOAD, JDLTYP, 
     3 ADLMAG, PREDEF, NPREDF, LFLAGS, MLVARX, DDLMAG, MDLOAD, PNEWDT, 
     4 JPROPS, NJPROP, PERIOD)
c
      INCLUDE 'ABA_PARAM.INC'
c     
      DIMENSION RHS(MLVARX,*),AMATRX(NDOFEL,NDOFEL),PROPS(*),
     1 SVARS(*),ENERGY(8),COORDS(MCRD,NNODE),U(NDOFEL),
     2 DU(MLVARX,*),V(NDOFEL),A(NDOFEL),TIME(2),PARAMS(*),
     3 JDLTYP(MDLOAD,*),ADLMAG(MDLOAD,*),DDLMAG(MDLOAD,*),
     4 PREDEF(2,NPREDF,NNODE),LFLAGS(*),JPROPS(*)
c
      DIMENSION ds1(3),ds2(3),dn(3),Trac(MCRD,NRHS),
     1 Trac_Jacob(MCRD,MCRD),R(MCRD,MCRD),coord_l(MCRD,NNODE),
     2 GP_coord(2),sf(3),B(MCRD,NDOFEL),co_de_m(3,3),
     3 B_t(NDOFEL,MCRD), Transformation_M(NDOFEL,NDOFEL),
     4 Transformation_M_T(NDOFEL,NDOFEL),temp1(MCRD,NDOFEL)
c
      DIMENSION stiff_l(NDOFEL,NDOFEL),temp2(NDOFEL,NDOFEL),
     1 stiff_g(NDOFEL,NDOFEL),residual_l(NDOFEL,NRHS),
     2 residual_g(NDOFEL,NRHS),aJacob_M(2,3),delu_loc_gp(mcrd),
     3 co_de(mcrd,nnode)
c
      DOUBLE PRECISION G_fn, G_ft, f_tn, f_tt, alpha, beta, rn, rt,
     1 p_m, p_n, deln, delt, tmax, pmax, opn, opt
c
c Define Inputs=======================================================
c 
      G_fn=PROPS(1)
      G_ft=PROPS(2)
      f_tn=PROPS(3)
      f_tt=PROPS(4)
      alpha=PROPS(5)
      beta=PROPS(6)
      rn=PROPS(7)
      rt=PROPS(8)
      GP_n=3d0
      GP_W = 1.0d0/3.0d0
c
c Initialize Matrices and Vectors=====================================
c 
      call k_vector_zero(ds1,3)
      call k_vector_zero(ds2,3)
      call k_vector_zero(dn,3)
      call k_matrix_zero(Trac,mcrd,nrhs)
      call k_matrix_zero(Trac_Jacob,mcrd,mcrd)
      call k_matrix_zero(R,mcrd,mcrd)
      call k_matrix_zero(coord_l,mcrd,nnode)
      call k_vector_zero(GP_coord,2)
      call k_vector_zero(sf,3)
      call k_matrix_zero(Transformation_M,ndofel,ndofel)
      call k_matrix_zero(Transformation_M_T,ndofel,ndofel)
      call k_matrix_zero(B,mcrd,ndofel)
      call k_matrix_zero(B_t,ndofel,mcrd)
      call k_matrix_zero(temp1,mcrd,ndofel)
      call k_matrix_zero(stiff_l,ndofel,ndofel)
      call k_matrix_zero(temp2,ndofel,ndofel)
      call k_matrix_zero(stiff_g,ndofel,ndofel)
      call k_matrix_zero(residual_l,ndofel,nrhs)
      call k_matrix_zero(residual_g,ndofel,nrhs)
      call k_matrix_zero(aJacob_M,2,3)
      call k_matrix_zero(rhs,ndofel,nrhs)
      call k_matrix_zero(amatrx,ndofel,ndofel)
      call k_matrix_zero(co_de,mcrd,nnode)
      a_Jacob=0.d0
c
c Do local computations===============================================
c 
      do i = 1, mcrd
         do j = 1, nnode
            co_de(i,j)=coords(i,j)+U(3.0*(j-1.0)+i)
         end do
      end do
c
      call k_local_coordinates(co_de,R,coord_l,Transformation_M,
     & Transformation_M_T,a_Jacob,aJacob_M,coords,u,ndofel,nnode,
     & mcrd)
c
c Compute shear and normal local opening displacments=================
c 
      do i = 1, 3
         ds1(i)=coord_l(1,i+3)-coord_l(1,i)
         ds2(i)=coord_l(2,i+3)-coord_l(2,i)
         dn(i) =coord_l(3,i+3)-coord_l(3,i)
      end do
c
c Do Calculations at Gauss Points=====================================
c
      do i = 1, GP_n
c
c Determine the values of the shape function at each Gauss Point
c
         call k_shape_fun(i,sf)
c
c Determine Inputs to Cohesive Model==================================
c
         p_m=(alpha*(alpha-1.0)*rn**2.0)/(1.0-alpha*rn**2.0)
         p_n=(beta*(beta-1.0)*rt**2.0)/(1.0-beta*rt**2.0)
         deln=(G_fn/f_tn)*alpha*rn*(1.0-rn)**(alpha-1.0)
     & *((alpha/p_m)+1.0)*((alpha/p_m)*rn+1.0)**(p_m-1.0)
         delt=(G_ft/f_tt)*beta*rt*(1.0-rt)**(beta-1.0)
     & *((beta/p_n)+1.0)*((beta/p_n)*rt+1.0)**(p_n-1.0)
c
         call k_vector_zero(delu_loc_gp,mcrd)
c
c Determine shear and normal opening displamenets at Gauss points
c         
         do j = 1, 3
            delu_loc_gp(1)=delu_loc_gp(1)+ds1(j)*sf(j)
            delu_loc_gp(2)=delu_loc_gp(2)+ds2(j)*sf(j)
            delu_loc_gp(3)=delu_loc_gp(3)+ dn(j)*sf(j)
         end do
c
         opn=delu_loc_gp(3)
         opt=sqrt(delu_loc_gp(1)**2.0+delu_loc_gp(2)**2.0)
c
         if ((Svars(GP_n*(i-1.0)+1.0) .LT. opt) .AND.
     & (opt .GT. rt*delt)) then
            Svars(GP_n*(i-1.0)+1.0)=opt
         end if
         if ((Svars(GP_n*(i-1.0)+2.0) .LT. opn) .AND.
     & (opn .GT. rn*deln)) then
            Svars(GP_n*(i-1.0)+2.0)=opn
         end if
         tmax=Svars(GP_n*(i-1.0)+1.0)
         pmax=Svars(GP_n*(i-1.0)+2.0)
c
c Determine Traction vector and tangent modulus matrix
c 
         call k_cohesive_law(Trac,Trac_Jacob,G_fn,G_ft,deln,delt,
     & alpha,beta,p_m,p_n,pmax,tmax,delu_loc_gp,mcrd,nrhs)
c
c Determine B matrix and its transpose
c 
         call k_Bmatrix(sf,B,mcrd,ndofel)
c
         call k_matrix_transpose(B,B_t,mcrd,ndofel)
c
c Compute the stiffness matrix
c Local Stiffness = B_t * Trac_Jacob * B
c
         call k_matrix_multiply(Trac_Jacob,B,temp1,mcrd,mcrd,
     & ndofel)
         call k_matrix_multiply(B_t,temp1,stiff_l,ndofel,
     & mcrd,ndofel)
c
c Compute Global stiffness matrix 
c Global_K = Transpose(T) * K * T
c
         call k_matrix_multiply(Transformation_M_T,stiff_l,
     & temp2,ndofel,ndofel,ndofel)
         call k_matrix_multiply(temp2,Transformation_M,stiff_g,
     & ndofel,ndofel,ndofel)
c
c Multiply Jacobian with the Global stiffness and add contribution
c from each Gauss Point
c
         a_Mult = a_Jacob*GP_w
         call k_matrix_plus_scalar(amatrx,stiff_g,a_Mult,
     & ndofel,ndofel)
c
c Compute the global residual vector
c Local_residual = B_t * Trac
c Global_residual = Transpose(T) * Local_residual
c 
         call k_matrix_multiply(B_t,Trac,residual_l,ndofel,
     & mcrd,nrhs)
         call k_matrix_multiply(Transformation_M_T,residual_l,
     & residual_g,ndofel,ndofel,nrhs)
c
c Multiply the Global residual by the Jacobian and add the 
c contribution from each point
c   
         call k_matrix_plus_scalar(rhs,residual_g,a_Mult,
     & ndofel,nrhs)
      end do
c
      return
      end
c=====================================================================
c=============================SUBROUTINES=============================
c=====================================================================
c
c Determine the global displacement-separation (B) matrix
c
      subroutine k_Bmatrix(sf,B,mcrd,ndofel)
      INCLUDE 'ABA_PARAM.INC'
      dimension sf(3),B(mcrd,ndofel)
      B(1,1) =  sf(1)
      B(1,4) =  sf(2)
      B(1,7) =  sf(3)
      B(1,10)= -sf(1)
      B(1,13)= -sf(2)
      B(1,16)= -sf(3)
      B(2,2) =  sf(1)
      B(2,5) =  sf(2)
      B(2,8) =  sf(3)
      B(2,11)= -sf(1)
      B(2,14)= -sf(2)
      B(2,17)= -sf(3)
      B(3,3) =  sf(1)
      B(3,6) =  sf(2)
      B(3,9) =  sf(3)
      B(3,12)= -sf(1)
      B(3,15)= -sf(2)
      B(3,18)= -sf(3)
c
      return
      end
c=====================================================================
      subroutine k_cohesive_law(T,T_d,G_fn,G_ft,deln,delt,
     & alpha,beta,p_m,p_n,pmax,tmax,delu,mcrd,nrhs)
      INCLUDE 'ABA_PARAM.INC'
      dimension T(mcrd,nrhs),T_d(mcrd,mcrd),delu(mcrd)
       DOUBLE PRECISION G_fn, G_ft, f_tn, f_tt, alpha, beta,
     & p_m, p_n, deln, delt, tmax, pmax, popn, popt, gam_n, 
     & gam_t, Tn, Tt, Dnn, Dnt, Dtt, T_d, T, delu 
c
      popn=delu(3)
      popt=sqrt(delu(1)**2.0+delu(2)**2.0)
c
      call k_Mac(pM1,G_fn,G_ft)
      call k_Mac(pM2,G_ft,G_fn)
c
      if (G_fn .NE. G_ft) then
         gam_n=(-G_fn)**(pM1/(G_fn-G_ft))*(alpha/p_m)**p_m
         gam_t=(-G_ft)**(pM2/(G_ft-G_fn))*(beta/p_n)**p_n
      elseif (G_fn .EQ. G_ft) then
         gam_n=-G_fn*(alpha/p_m)**p_m
         gam_t=(beta/p_n)**p_n
      end if
c
c Pre-calculation of the normal cohesive traction Tn
c
      if (popn .LT. 0.0) then
         popn=0.0
      elseif ((popn .GE. deln) .OR. (popt .GE. delt)) then
         Tn = 0.0
      elseif (popn .GE. pmax) then
         Tn=(gam_n/deln)*(p_m*(1.0-(popn/deln))**alpha*((p_m/alpha)
     & +(popn/deln))**(p_m-1.0)-alpha*((1.0-(popn/deln))**(alpha-1.0))
     & *((p_m/alpha)+(popn/deln))**p_m)*(gam_t*(1.0-(popt/delt))
     & **beta*((p_n/beta)+(popt/delt))**p_n+pM2)
      else
         Tn=(gam_n/deln)*(p_m*(1.0-(pmax/deln))**alpha*((p_m/alpha)
     & +(pmax/deln))**(p_m-1.0)-alpha*((1.0-(pmax/deln))**(alpha-1.0))
     & *((p_m/alpha)+(pmax/deln))**p_m)*(gam_t*(1.0-(popt/delt))
     & **beta*((p_n/beta)+(popt/delt))**p_n+pM2)
     & *popn/pmax
      end if
c
c Pre-calculation of the tangential cohesive traction, Tt
c
      if ((popn .GE. deln) .OR. (popt .GE. delt)) then
         Tt = 0.0
      elseif (popt .GE. tmax) then
         Tt=(gam_t/delt)*(p_n*(1.0-(popt/delt))**beta*((p_n/beta)
     & +(popt/delt))**(p_n-1.0)-beta*((1.0-(popt/delt))**(beta-1.0))
     & *((p_n/beta)+(popt/delt))**p_n)*(gam_n*(1.0-(popn/deln))
     & **alpha*((p_m/alpha)+(popn/deln))**p_m+pM1) 
      else
         Tt=(gam_t/delt)*(p_n*(1.0-(tmax/delt))**beta*((p_n/beta)
     & +(tmax/delt))**(p_n-1.0)-beta*((1.0-(tmax/delt))**(beta-1.0))
     & *((p_n/beta)+(tmax/delt))**p_n)*(gam_n*(1.0-(popn/deln))
     & **alpha*((p_m/alpha)+(popn/deln))**p_m+pM1)
     & *popt/tmax     
      end if 
c
c Algortihm 1 
c
c Normal Cohesive Interaction
c (1) Contact
      if (delu(3) .LT. 0.0) then
         Dnn = -(gam_n/(deln**2))*(p_m/alpha)**(p_m-1.0)*(alpha+p_m)*
     & (gam_t*(p_n/beta)**p_n + pM2)
         Dnt = 0.0
         Tn = Dnn * delu(3)
      else if ((popn .LT. deln) .AND. (popt .LT. delt) 
     & .AND. (Tn .GE. -1.0E-5)) then
         Tn = Tn
c (2) Softening Condition
         if (popn .GE. pmax) then
         Dnn=(gam_n/(deln**2.0))*((p_m**2.0-p_m)*((1.0-(popn/deln))
     & **alpha)*((p_m/alpha)+(popn/deln))**(p_m-2.0)+(alpha**2.0
     & -alpha)*((1.0-(popn/deln))**(alpha-2.0))*((p_m/alpha)
     & +(popn/deln))**p_m -2.0*alpha*p_m*((1.0-(popn/deln))
     & **(alpha-1.0))*((p_m/alpha)+(popn/deln))**(p_m-1.0))*(gam_t*
     & (1.0-(popt/delt))**beta*((p_n/beta)+(popt/delt))**p_n+pM2)
         Dnt=(gam_n*gam_t/(deln*delt))*(p_m*((1.0-(popn/deln))**alpha)
     & *((p_m/alpha)+(popn/deln))**(p_m-1.0)-alpha*((1.0-(popn/deln))
     & **(alpha-1.0))*((p_m/alpha)+(popn/deln))**p_m)*(p_n
     & *((1.0-(popt/delt))**beta)*(((p_n/beta)+(popt/delt))
     & **(p_n-1.0))-beta*((1.0-(popt/delt))**(beta-1.0))*((p_n/beta)
     & +(popt/delt))**p_n)
c (3) Unloading/reloading condition
         else
         Dnn=(gam_n/deln)*(p_m*(1.0-(pmax/deln))**alpha*((p_m/alpha)
     & +(pmax/deln))**(p_m-1.0)-alpha*((1.0-(pmax/deln))**(alpha-1.0))
     & *((p_m/alpha)+(pmax/deln))**p_m)*(gam_t*(1.0-(popt/delt))
     & **beta*((p_n/beta)+(popt/delt))**p_n+pM2)
     & /pmax
         Dnt=(gam_n*gam_t/(deln*delt))*(p_m*((1.0-(pmax/deln))**alpha)
     & *((p_m/alpha)+(pmax/deln))**(p_m-1.0)-alpha*((1.0-(pmax/deln))
     & **(alpha-1.0))*((p_m/alpha)+(pmax/deln))**p_m)*(p_n
     & *((1.0-(popt/delt))**beta)*(((p_n/beta)+(popt/delt))
     & **(p_n-1.0))-beta*((1.0-(popt/delt))**(beta-1.0))*((p_n/beta)
     & +(popt/delt))**p_n)
     & *popn/pmax
         end if
c (4) Complete Failure
      else
         Tn = 0.0
         Dnn = 0.0
         Dnt = 0.0
      end if    
c
c Tangential Cohesive Interaction
c
      if ((popt .LT. delt) .AND. (popn .LT. deln) 
     & .AND. (Tt .GE. -1.0E-5)) then
        Tt = Tt
c (1) Softening Condition
        if (popt .GE. tmax) then
           Dtt=(gam_t/(delt**2.0))*((p_n**2.0-p_n)*((1.0-(popt/delt))
     & **beta)*((p_n/beta)+(popt/delt))**(p_n-2.0)+(beta**2.0-beta)
     & *((1.0-(popt/delt))**(beta-2.0))*((p_n/beta)+(popt/delt))**p_n
     & -2.0*beta*p_n*((1.0-(popt/delt))**(beta-1.0))*((p_n/beta)
     & +(popt/delt))**(p_n-1.0))*(gam_n*(1.0-(popn/deln))
     & **alpha*((p_m/alpha)+(popn/deln))**p_m+pM1)
           Dtn=(gam_n*gam_t/(deln*delt))*(p_m*((1.0-(popn/deln))
     & **alpha)*((p_m/alpha)+(popn/deln))**(p_m-1.0)-alpha*((1.0-
     & (popn/deln))**(alpha-1.0))*((p_m/alpha)+(popn/deln))**p_m)*(p_n
     & *((1.0-(popt/delt))**beta)*(((p_n/beta)+(popt/delt))
     & **(p_n-1.0))-beta*((1.0-(popt/delt))**(beta-1.0))*((p_n/beta)
     & +(popt/delt))**p_n)
c (2) Unloading/reloading condition
        else
           Dtt=(gam_t/delt)*(p_n*(1.0-(tmax/delt))**beta*((p_n/beta)
     & +(tmax/delt))**(p_n-1.0)-beta*((1.0-(tmax/delt))**(beta-1.0))
     & *((p_n/beta)+(tmax/delt))**p_n)*(gam_n*(1.0-(popn/deln))
     & **alpha*((p_m/alpha)+(popn/deln))**p_m+pM1)
     & /tmax 
           Dtn=(gam_n*gam_t/(deln*delt))*(p_m*((1.0-(popn/deln))
     & **alpha)*((p_m/alpha)+(popn/deln))**(p_m-1.0)-alpha*((1.0-
     & (popn/deln))**(alpha-1.0))*((p_m/alpha)+(popn/deln))**p_m)*(p_n
     & *((1.0-(tmax/delt))**beta)*(((p_n/beta)+(tmax/delt))
     & **(p_n-1.0))-beta*((1.0-(tmax/delt))**(beta-1.0))*((p_n/beta)+
     & (tmax/delt))**p_n)
     & *popt/tmax
        end if
c (3) Complete failure condition
      else
        Tt = 0.0
        Dtt = 0.0
        Dtn = 0.0
      end if
      if (Dtn .NE. Dnt) then
        Dtn = 0.5*(Dtn + Dnt)
        Dnt = Dtn
      end if
c
      if (popt .EQ. 0.0) then
         T(1,1) = 0.0
         T(2,1) = 0.0
         T(3,1) = Tn
c
         T_d(1,1)=Dtt
         T_d(1,2)=Dtn
         T_d(1,3)=0.0
         T_d(2,1)=Dnt
         T_d(2,2)=Dtt
         T_d(2,3)=0.0
         T_d(3,1)=0.0
         T_d(3,2)=0.0
         T_d(3,3)=Dnn
      else
         T(1,1)=Tt*delu(1)/popt
         T(2,1)=Tt*delu(2)/popt
         T(3,1)=Tn
c
         T_d(1,1)=Dtt*(delu(1)/popt)**2.0+Tt*((delu(2)**2.0)
     &    /(popt**3.0))
         T_d(1,2)=Dtt*delu(1)*delu(2)/(popt**2.0)
     &    -Tt*delu(1)*delu(2)/(popt**3.0)
         T_d(1,3)=Dnt*delu(1)/popt
c
         T_d(2,1)=Dtt*delu(1)*delu(2)/(popt**2.0)
     &    -Tt*delu(1)*delu(2)/(popt**3.0)
         T_d(2,2)=Dtt*(delu(2)/popt)**2.0+Tt*((delu(1)**2.0)
     &    /(popt**3.0))
         T_d(2,3)=Dnt*delu(2)/popt
c
         T_d(3,1)=Dtn*delu(1)/popt
         T_d(3,2)=Dtn*delu(2)/popt
         T_d(3,3)=Dnn
      end if
c
      return
      end
c=====================================================================
      subroutine k_local_coordinates(co_de,R,coord_l,Transformation_M,
     & Transformation_M_T,a_Jacob,aJacob_M,coords,u,ndofel,nnode,
     & mcrd)
      INCLUDE 'ABA_PARAM.INC'
      dimension R(mcrd,mcrd),coord_l(mcrd,nnode),aJacob_M(2,3),
     & Transformation_M(ndofel,ndofel),coords(mcrd,nnode),
     & Transformation_M_T(ndofel,ndofel),u(ndofel),
     & co_de(mcrd,nnode), co_de_m(3,3),SFD(2,4)
c
      call k_matrix_zero(co_de_m,3,4)
c
      do i = 1, 3
         co_de_m(i,1)=(co_de(i,1)+co_de(i,4))*0.5
         co_de_m(i,2)=(co_de(i,2)+co_de(i,5))*0.5
         co_de_m(i,3)=(co_de(i,3)+co_de(i,6))*0.5
      end do
c
      SFD(1,1) =-1
      SFD(1,2) = 1
      SFD(1,3) = 0
      SFD(2,1) =-1
      SFD(2,2) = 0
      SFD(2,3) = 1
c
      do i = 1,2
         do j = 1,3
            do k =1, 3
               aJacob_M(i,j) = aJacob_M(i,j) + SFD(i,k)*co_de_m(j,k)
            end do
         end do
      end do

      dum1 = aJacob_M(1,2)*aJacob_M(2,3) - aJacob_M(1,3)*aJacob_M(2,2)
      dum2 = aJacob_M(1,3)*aJacob_M(2,1) - aJacob_M(1,1)*aJacob_M(2,3)
      dum3 = aJacob_M(1,1)*aJacob_M(2,2) - aJacob_M(1,2)*aJacob_M(2,1)
c
      a_Jacob = sqrt(dum1**2 + dum2**2 + dum3**2)/2.0d0
      Rn1 = sqrt(dum1**2 + dum2**2 + dum3**2)
c
      R(3,1) = dum1/Rn1
      R(3,2) = dum2/Rn1
      R(3,3) = dum3/Rn1
c
      aLen=sqrt(aJacob_M(1,1)**2.0 + aJacob_M(1,2)**2.0
     & + aJacob_M(1,3)**2.0)
      R(1,1)=aJacob_M(1,1)/aLen
      R(1,2)=aJacob_M(1,2)/aLen
      R(1,3)=aJacob_M(1,3)/aLen
c
      R(2,1)=R(3,2)*R(1,3)-R(3,3)*R(1,2)
      R(2,2)=R(3,3)*R(1,1)-R(3,1)*R(1,3)
      R(2,3)=R(3,1)*R(1,2)-R(3,2)*R(1,1)
c
c=====================================================================
      num=nnode
c
      do i = 1, num
         dum=3.0*(i-1.0)
         Transformation_M(dum+1,dum+1)=R(1,1)
         Transformation_M(dum+1,dum+2)=R(1,2) 
         Transformation_M(dum+1,dum+3)=R(1,3)
         Transformation_M(dum+2,dum+1)=R(2,1)
         Transformation_M(dum+2,dum+2)=R(2,2)
         Transformation_M(dum+2,dum+3)=R(2,3)
         Transformation_M(dum+3,dum+1)=R(3,1)
         Transformation_M(dum+3,dum+2)=R(3,2)
         Transformation_M(dum+3,dum+3)=R(3,3)
      end do
c
      call k_matrix_transpose(Transformation_M,Transformation_M_T,
     $ ndofel,ndofel)
c
      do i = 1, nnode
         coord_l(1,i)=(R(1,1)*co_de(1,i)+R(1,2)*co_de(2,i)
     & +R(1,3)*co_de(3,i))
         coord_l(2,i)=(R(2,1)*co_de(1,i)+R(2,2)*co_de(2,i)
     & +R(2,3)*co_de(3,i))
         coord_l(3,i)=(R(3,1)*co_de(1,i)+R(3,2)*co_de(2,i)
     & +R(3,3)*co_de(3,i))
      end do
c    
      return
      end
c=====================================================================
      subroutine k_shape_fun(i,sf)
      INCLUDE 'ABA_PARAM.INC'
      dimension sf(3), GP_coord(2)
c
      if (i .eq. 1) then
         GP_coord(1)= 1.0d0/6.0d0
         GP_coord(2)= 2.0d0/3.0d0
      elseif (i .eq. 2) then
         GP_coord(1)= 2.0d0/3.0d0
         GP_coord(2)= 1.0d0/6.0d0
      elseif (i .eq. 3) then
         GP_coord(1)= 1.0d0/6.0d0
         GP_coord(2)= 1.0d0/6.0d0
      end if
c
      sf(1)= 1.0 - GP_coord(1) - GP_coord(2)
      sf(2)= GP_coord(1)
      sf(3)= GP_coord(2)
c
      return
      end
c=====================================================================
      subroutine k_matrix_multiply(A,B,C,l,n,m)
      INCLUDE 'ABA_PARAM.INC'
      dimension A(l,n),B(n,m),C(l,m)
c
      call k_matrix_zero(C,l,m)
c
      do i = 1, l
         do j = 1, m
            do k = 1, n
               C(i,j)=C(i,j)+A(i,k)*B(k,j)
            end do
         end do
      end do
c
      return
      end
c=====================================================================
      subroutine k_matrix_plus_scalar(A,B,c,n,m)
      INCLUDE 'ABA_PARAM.INC'
      dimension A(n,m),B(n,m)
c
      do i = 1, n
         do j = 1, m
            A(i,j)=A(i,j)+c*B(i,j)
         end do
      end do
c
      return
      end
c=====================================================================
      subroutine k_matrix_transpose(A,B,n,m)
      INCLUDE 'ABA_PARAM.INC'
      dimension A(n,m),B(m,n)
c
      do i = 1, n
         do j = 1, m
            B(j,i)=A(i,j)
         end do
      end do
c
      return
      end
c=====================================================================
      subroutine k_matrix_zero(A,n,m)
      INCLUDE 'ABA_PARAM.INC'
      dimension A(n,m)
c
      do i = 1, n
         do j = 1, m
            A(i,j)=0.d0
         end do
      end do
c
      return
      end
c=====================================================================
      subroutine k_vector_zero(A,n)
      INCLUDE 'ABA_PARAM.INC'
      dimension A(n)
c
      do i = 1, n
         A(i)=0.d0
      end do
c
      return
      end
c=====================================================================
      subroutine k_Mac(pM,a,b)
      INCLUDE 'ABA_PARAM.INC'
c
      if ((a-b) .GE. 0.0) then
         pM=a-b
      elseif ((a-b) .LT. 0.0) then
         pM=0.d0
      end if
c
      return
      end
c=====================================================================
c=================================END=================================
c=====================================================================









      
